#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <math.h>
#include <dos.h>
#include <mem.h>

#ifndef uchar
#	define uchar unsigned char
#endif

enum {
	CMD_NONE = 0,
	CMD_F_UP,
	CMD_F_DN,
	CMD_I_UP,
	CMD_I_DN
};

#define D_SIZE 5	/* rozmiar diody */
#define SH_SIZE 64	/* rozmiar cienia */
#define DELAY 10	/* opoznienie petli w ms */

#define initmath() pi = 4 * atan(1); pi2 = 2 * pi
#define PIXEL(x,y) (*(char *) (0xA0000 + ((y) << 8) + ((y) << 6) + (x)))

static double pi, pi2;
double angle[3];	/* katy poszczegolnych skladowych */
int speed[3];	/* predkosci zmian skladowych (max 1023) */
int intens[3] = {127, 127, 127};	/* intensywnosc skladowych (max 127) */
int rgb[3];	/* aktualne wartosci skladowych rgb */

void setreg (int reg, int *trgb)
{
	int i;

	outp (0x3C8, reg);

	for (i = 0; i < 3; i++)
		outp (0x3C9, trgb[i]);
}

void setrgb (void)
{
	int i, trgb[3];

	memcpy (trgb, rgb, sizeof(trgb));
	for (i = 0; i < 3; i++) {
		trgb[i] *= intens[i];
		trgb[i] /= 127;
	}

	for (i = 0; i <= SH_SIZE; i++) {
		if (trgb[0])
			trgb[0]--;
		if (trgb[1])
			trgb[1]--;
		if (trgb[2])
			trgb[2]--;
		setreg (i + 1, &trgb);
	}
}

void fill (int x, int y, uchar cfill)
{
	if (x < 0 || y < 0 || x > 319 || y > 199)
		return;

	PIXEL(x, y) = cfill;

	if (PIXEL(x - 1, y) != cfill)
		fill (x - 1, y, cfill);
	if (PIXEL(x + 1, y) != cfill)
		fill (x + 1, y, cfill);
	if (PIXEL(x, y - 1) != cfill)
		fill (x, y - 1, cfill);
	if (PIXEL(x, y + 1) != cfill)
		fill (x, y + 1, cfill);
}

void shade (int x, int y, int size, uchar color)
{
	int tx, ty;

	x -= (size >> 1);
	y -= (size >> 1);

	if (x < 0 || y < 0 || (x + size) >= 319 || (y + size) >= 199)
		return;

	for (ty = 0; ty < size; ty++)
		for (tx = 0; tx < size; tx++) {
			int bx, by;

			if (PIXEL(x + tx, y + ty) != color - 1)
				continue;

			for (by = -1; by <= 1; by++)
				for (bx = -1; bx <= 1; bx++)
					if (!PIXEL(x + tx + bx, y + ty + by))
						PIXEL(x + tx + bx, y + ty + by) = color;
		}
}

void update_rgb (void)
{
	int i;

	for (i = 0; i < 3; i++) {
		angle[i] += (double) speed[i] / 1024;
		while (angle[i] >= pi2)
			angle[i] -= pi2;
		if (angle[i] < 0)
			angle[i] += pi2;

		rgb[i] = sin(angle[i]) * 31 + 32;
	}

	setrgb();
}

void initgraph (void)
{
	int i;

	_asm {
		mov ax, 13h
		int 10h
	}

	update_rgb();

	for (i = 0; i < 1024; i++) {
		int x, y;
		double a;

		a = (i * pi2) / 1024;
		x = cos(a) * D_SIZE + 160;
		y = sin(a) * D_SIZE + 100;

		PIXEL(x, y) = 1;
	}

	fill (160, 100, 1);
	for (i = 0; i < SH_SIZE; i++)
		shade (160, 100, (D_SIZE + i) * 2 + 2, i + 2);
}

void freegraph (void)
{
	_asm {
		mov ax, 3h
		int 10h
	}
}

void initmisc (void)
{
	int i;

	srand (time(NULL));

	for (i = 0; i < 3; i++) {
		speed[i] = (rand() & 31) + 64; // rand() & 1023;
		angle[i] = (double) (rand() % (int) (16384 * pi2)) / 16384;
	}
}

/*
	q, a - zmiana predkosci r
	w, s - zmiana predkosci g
	e, d - zmiana predkosci b

	r, f - zmiana intensywnosci r
	t, g - zmiana intensywnosci g
	y, h - zmiana intensywnosci b

	esc - wyjscie (obsluga w main)
*/

void do_cmd (int ch)
{
	struct keys_t {
		char key;
		int index;
		char cmd;
	} static keys[] = {
		'q', 0, CMD_F_UP,
		'a', 0, CMD_F_DN,
		'r', 0, CMD_I_UP,
		'f', 0, CMD_I_DN,

		'w', 1, CMD_F_UP,
		's', 1, CMD_F_DN,
		't', 1, CMD_I_UP,
		'g', 1, CMD_I_DN,

		'e', 2, CMD_F_UP,
		'd', 2, CMD_F_DN,
		'y', 2, CMD_I_UP,
		'h', 2, CMD_I_DN
	};

	int index, i;
	char cmd = CMD_NONE;

	for (i = 0; i < sizeof(keys) / sizeof(struct keys_t); i++) {
		if (ch == keys[i].key) {
			cmd = keys[i].cmd;
			index = keys[i].index;
			break;
		}
	}

	switch (cmd) {
		case CMD_F_UP:
			if (speed[index] < 1023)
				speed[index]++;
			break;
		case CMD_F_DN:
			if (speed[index])
				speed[index]--;
			break;
		case CMD_I_UP:
			if (intens[index] < 63) {
				intens[index]++;
				setrgb();
			}
			break;
		case CMD_I_DN:
			if (intens[index]) {
				intens[index]--;
				setrgb();
			}
			break;
		default:
			break;
	}
}

int main (int ac, char **av)
{
	int ch = 0;

	initmath();
	initmisc();
	initgraph();

	while (ch != 27) {	/* esc */
		if (kbhit()) {
			ch = getch();
			if (!ch)	/* kursory itp. */
				ch = getch() | 0x100;
			do_cmd (ch);
		} else	/* jakby DELAY > rate klawiatury */
			delay (DELAY);

		update_rgb();
	}

	freegraph();

	return 0;
}

